@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004, 2005, 2007-2009, 2012  Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2004, 2005, 2007-2009, 2012-2015  Internet Systems Consortium, Inc. ("ISC")
  * Copyright (C) 1999-2002  Internet Software Consortium.
  *
  * Permission to use, copy, modify, and/or distribute this software for any
@@ -15,7 +15,7 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* $Id$ */
+/* $Id: os.c,v 1.39 2012/02/06 23:46:44 tbox Exp $ */
 
 #include <config.h>
 #include <stdarg.h>
@@ -56,7 +56,7 @@ static char *version_error =
 	"named requires Windows 2000 Service Pack 2 or later to run correctly";
 
 void
-ns_paths_init() {
+ns_paths_init(void) {
 	if (!Initialized)
 		isc_ntpaths_init();
 
@@ -80,9 +80,10 @@ ns_paths_init() {
 static void
 version_check(const char *progname) {
 
-	if(isc_win32os_majorversion() < 5)
+	if ((isc_win32os_versioncheck(4, 0, 0, 0) >= 0) &&
+	    (isc_win32os_versioncheck(5, 0, 0, 0) < 0))
 		return;	/* No problem with Version 4.0 */
-	if(isc_win32os_versioncheck(5, 0, 2, 0) < 0)
+	if (isc_win32os_versioncheck(5, 0, 2, 0) < 0)
 		if (ntservice_isservice())
 			NTReportError(progname, version_error);
 		else
@@ -293,7 +294,7 @@ isc_result_t
 ns_os_gethostname(char *buf, size_t len) {
 	int n;
 
-	n = gethostname(buf, len);
+	n = gethostname(buf, (int)len);
 	return ((n == 0) ? ISC_R_SUCCESS : ISC_R_FAILURE);
 }
 
