@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2006-2008  Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2006-2008, 2010-2015  Internet Systems Consortium, Inc. ("ISC")
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -14,7 +14,7 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* $Id: httpd.c,v 1.12.12.3 2008/08/08 05:10:34 marka Exp $ */
+/* $Id$ */
 
 /*! \file */
 
@@ -26,6 +26,7 @@
 #include <isc/socket.h>
 #include <isc/string.h>
 #include <isc/task.h>
+#include <isc/time.h>
 #include <isc/util.h>
 
 #include <string.h>
@@ -58,20 +59,6 @@
 #define HTTP_SENDGROW			1024
 #define HTTP_SEND_MAXLEN		10240
 
-/*%
- * HTTP urls.  These are the URLs we manage, and the function to call to
- * provide the data for it.  We pass in the base url (so the same function
- * can handle multiple requests), and a structure to fill in to return a
- * result to the client.  We also pass in a pointer to be filled in for
- * the data cleanup function.
- */
-struct isc_httpdurl {
-	char			       *url;
-	isc_httpdaction_t	       *action;
-	void			       *action_arg;
-	ISC_LINK(isc_httpdurl_t)	link;
-};
-
 #define HTTPD_CLOSE		0x0001 /* Got a Connection: close header */
 #define HTTPD_FOUNDHOST		0x0002 /* Got a Host: header */
 
@@ -87,6 +74,7 @@ struct isc_httpd {
 	 */
 	char			recvbuf[HTTP_RECVLEN]; /*%< receive buffer */
 	isc_uint32_t		recvlen;	/*%< length recv'd */
+	char		       *headers;	/*%< set in process_request() */
 	unsigned int		method;
 	char		       *url;
 	char		       *querystring;
@@ -151,6 +139,7 @@ struct isc_httpdmgr {
 
 	ISC_LIST(isc_httpdurl_t) urls;		/*%< urls we manage */
 	isc_httpdaction_t      *render_404;
+	isc_httpdaction_t      *render_500;
 };
 
 /*%
@@ -216,15 +205,12 @@ static isc_result_t process_request(isc_httpd_t *, int);
 static void httpdmgr_destroy(isc_httpdmgr_t *);
 static isc_result_t grow_headerspace(isc_httpd_t *);
 static void reset_client(isc_httpd_t *httpd);
-static isc_result_t render_404(const char *, const char *,
-			       void *,
-			       unsigned int *, const char **,
-			       const char **, isc_buffer_t *,
-			       isc_httpdfree_t **, void **);
+
+static isc_httpdaction_t render_404;
+static isc_httpdaction_t render_500;
 
 static void
-destroy_client(isc_httpd_t **httpdp)
-{
+destroy_client(isc_httpd_t **httpdp) {
 	isc_httpd_t *httpd = *httpdp;
 	isc_httpdmgr_t *httpdmgr = httpd->mgr;
 
@@ -300,6 +286,7 @@ isc_httpdmgr_create(isc_mem_t *mctx, isc_socket_t *sock, isc_task_t *task,
 		goto cleanup;
 
 	httpd->render_404 = render_404;
+	httpd->render_500 = render_500;
 
 	*httpdp = httpd;
 	return (ISC_R_SUCCESS);
@@ -308,14 +295,13 @@ isc_httpdmgr_create(isc_mem_t *mctx, isc_socket_t *sock, isc_task_t *task,
 	isc_task_detach(&httpd->task);
 	isc_socket_detach(&httpd->sock);
 	isc_mem_detach(&httpd->mctx);
-	isc_mutex_destroy(&httpd->lock);
+	(void)isc_mutex_destroy(&httpd->lock);
 	isc_mem_put(mctx, httpd, sizeof(isc_httpdmgr_t));
 	return (result);
 }
 
 static void
-httpdmgr_destroy(isc_httpdmgr_t *httpdmgr)
-{
+httpdmgr_destroy(isc_httpdmgr_t *httpdmgr) {
 	isc_mem_t *mctx;
 	isc_httpdurl_t *url;
 
@@ -357,7 +343,7 @@ httpdmgr_destroy(isc_httpdmgr_t *httpdmgr)
 	}
 
 	UNLOCK(&httpdmgr->lock);
-	isc_mutex_destroy(&httpdmgr->lock);
+	(void)isc_mutex_destroy(&httpdmgr->lock);
 
 	if (httpdmgr->ondestroy != NULL)
 		(httpdmgr->ondestroy)(httpdmgr->cb_arg);
@@ -372,8 +358,7 @@ httpdmgr_destroy(isc_httpdmgr_t *httpdmgr)
 #define BUFLENOK(s) (httpd->recvbuf - (s) < HTTP_RECVLEN)
 
 static isc_result_t
-process_request(isc_httpd_t *httpd, int length)
-{
+process_request(isc_httpd_t *httpd, int length) {
 	char *s;
 	char *p;
 	int delim;
@@ -383,6 +368,7 @@ process_request(isc_httpd_t *httpd, int length)
 	httpd->recvlen += length;
 
 	httpd->recvbuf[httpd->recvlen] = 0;
+	httpd->headers = NULL;
 
 	/*
 	 * If we don't find a blank line in our buffer, return that we need
@@ -487,6 +473,8 @@ process_request(isc_httpd_t *httpd, int length)
 	p = s + 1;
 	s = p;
 
+	httpd->headers = s;
+
 	if (strstr(s, "Connection: close") != NULL)
 		httpd->flags |= HTTPD_CLOSE;
 
@@ -506,8 +494,7 @@ process_request(isc_httpd_t *httpd, int length)
 }
 
 static void
-isc_httpd_accept(isc_task_t *task, isc_event_t *ev)
-{
+isc_httpd_accept(isc_task_t *task, isc_event_t *ev) {
 	isc_result_t result;
 	isc_httpdmgr_t *httpdmgr = ev->ev_arg;
 	isc_httpd_t *httpd;
@@ -579,6 +566,8 @@ isc_httpd_accept(isc_task_t *task, isc_event_t *ev)
 	r.length = HTTP_RECVLEN - 1;
 	result = isc_socket_recv(httpd->sock, &r, 1, task, isc_httpd_recvdone,
 				 httpd);
+	/* FIXME!!! */
+	POST(result);
 	NOTICE("accept queued recv on socket");
 
  requeue:
@@ -600,8 +589,8 @@ isc_httpd_accept(isc_task_t *task, isc_event_t *ev)
 }
 
 static isc_result_t
-render_404(const char *url, const char *querystring,
-	   void *arg,
+render_404(const char *url, isc_httpdurl_t *urlinfo,
+	   const char *querystring, const char *headers, void *arg,
 	   unsigned int *retcode, const char **retmsg,
 	   const char **mimetype, isc_buffer_t *b,
 	   isc_httpdfree_t **freecb, void **freecb_args)
@@ -609,7 +598,9 @@ render_404(const char *url, const char *querystring,
 	static char msg[] = "No such URL.";
 
 	UNUSED(url);
+	UNUSED(urlinfo);
 	UNUSED(querystring);
+	UNUSED(headers);
 	UNUSED(arg);
 
 	*retcode = 404;
@@ -623,16 +614,41 @@ render_404(const char *url, const char *querystring,
 	return (ISC_R_SUCCESS);
 }
 
-static void
-isc_httpd_recvdone(isc_task_t *task, isc_event_t *ev)
+static isc_result_t
+render_500(const char *url, isc_httpdurl_t *urlinfo,
+	   const char *querystring, const char *headers, void *arg,
+	   unsigned int *retcode, const char **retmsg,
+	   const char **mimetype, isc_buffer_t *b,
+	   isc_httpdfree_t **freecb, void **freecb_args)
 {
+	static char msg[] = "Internal server failure.";
+
+	UNUSED(url);
+	UNUSED(urlinfo);
+	UNUSED(querystring);
+	UNUSED(headers);
+	UNUSED(arg);
+
+	*retcode = 500;
+	*retmsg = "Internal server failure";
+	*mimetype = "text/plain";
+	isc_buffer_reinit(b, msg, strlen(msg));
+	isc_buffer_add(b, strlen(msg));
+	*freecb = NULL;
+	*freecb_args = NULL;
+
+	return (ISC_R_SUCCESS);
+}
+
+static void
+isc_httpd_recvdone(isc_task_t *task, isc_event_t *ev) {
 	isc_region_t r;
 	isc_result_t result;
 	isc_httpd_t *httpd = ev->ev_arg;
 	isc_socketevent_t *sev = (isc_socketevent_t *)ev;
 	isc_httpdurl_t *url;
 	isc_time_t now;
-	char datebuf[32];  /* Only need 30, but safety first */
+	char datebuf[ISC_FORMATHTTPTIMESTAMP_SIZE];
 
 	ENTER("recv");
 
@@ -652,8 +668,9 @@ isc_httpd_recvdone(isc_task_t *task, isc_event_t *ev)
 		}
 		r.base = (unsigned char *)httpd->recvbuf + httpd->recvlen;
 		r.length = HTTP_RECVLEN - httpd->recvlen - 1;
-		result = isc_socket_recv(httpd->sock, &r, 1, task,
-					 isc_httpd_recvdone, httpd);
+		/* check return code? */
+		(void)isc_socket_recv(httpd->sock, &r, 1, task,
+				      isc_httpd_recvdone, httpd);
 		goto out;
 	} else if (result != ISC_R_SUCCESS) {
 		destroy_client(&httpd);
@@ -676,8 +693,9 @@ isc_httpd_recvdone(isc_task_t *task, isc_event_t *ev)
 		url = ISC_LIST_NEXT(url, link);
 	}
 	if (url == NULL)
-		result = httpd->mgr->render_404(httpd->url, httpd->querystring,
-						NULL,
+		result = httpd->mgr->render_404(httpd->url, NULL,
+						httpd->querystring,
+						NULL, NULL,
 						&httpd->retcode,
 						&httpd->retmsg,
 						&httpd->mimetype,
@@ -685,23 +703,43 @@ isc_httpd_recvdone(isc_task_t *task, isc_event_t *ev)
 						&httpd->freecb,
 						&httpd->freecb_arg);
 	else
-		result = url->action(httpd->url, httpd->querystring,
+		result = url->action(httpd->url, url,
+				     httpd->querystring,
+				     httpd->headers,
 				     url->action_arg,
 				     &httpd->retcode, &httpd->retmsg,
 				     &httpd->mimetype, &httpd->bodybuffer,
 				     &httpd->freecb, &httpd->freecb_arg);
 	if (result != ISC_R_SUCCESS) {
-		destroy_client(&httpd);
-		goto out;
+		result = httpd->mgr->render_500(httpd->url, url,
+						httpd->querystring,
+						NULL, NULL,
+						&httpd->retcode,
+						&httpd->retmsg,
+						&httpd->mimetype,
+						&httpd->bodybuffer,
+						&httpd->freecb,
+						&httpd->freecb_arg);
+		RUNTIME_CHECK(result == ISC_R_SUCCESS);
 	}
 
 	isc_httpd_response(httpd);
 	isc_httpd_addheader(httpd, "Content-Type", httpd->mimetype);
 	isc_httpd_addheader(httpd, "Date", datebuf);
 	isc_httpd_addheader(httpd, "Expires", datebuf);
-	isc_httpd_addheader(httpd, "Last-Modified", datebuf);
-	isc_httpd_addheader(httpd, "Pragma: no-cache", NULL);
-	isc_httpd_addheader(httpd, "Cache-Control: no-cache", NULL);
+
+	if (url != NULL && url->isstatic) {
+		char loadbuf[ISC_FORMATHTTPTIMESTAMP_SIZE];
+		isc_time_formathttptimestamp(&url->loadtime,
+					     loadbuf, sizeof(loadbuf));
+		isc_httpd_addheader(httpd, "Last-Modified", loadbuf);
+		isc_httpd_addheader(httpd, "Cache-Control: public", NULL);
+	} else {
+		isc_httpd_addheader(httpd, "Last-Modified", datebuf);
+		isc_httpd_addheader(httpd, "Pragma: no-cache", NULL);
+		isc_httpd_addheader(httpd, "Cache-Control: no-cache", NULL);
+	}
+
 	isc_httpd_addheader(httpd, "Server: libisc", NULL);
 	isc_httpd_addheaderuint(httpd, "Content-Length",
 				isc_buffer_usedlength(&httpd->bodybuffer));
@@ -716,8 +754,9 @@ isc_httpd_recvdone(isc_task_t *task, isc_event_t *ev)
 	if (isc_buffer_length(&httpd->bodybuffer) > 0)
 		ISC_LIST_APPEND(httpd->bufflist, &httpd->bodybuffer, link);
 
-	result = isc_socket_sendv(httpd->sock, &httpd->bufflist, task,
-				  isc_httpd_senddone, httpd);
+	/* check return code? */
+	(void)isc_socket_sendv(httpd->sock, &httpd->bufflist, task,
+			       isc_httpd_senddone, httpd);
 
  out:
 	isc_event_free(&ev);
@@ -725,8 +764,7 @@ isc_httpd_recvdone(isc_task_t *task, isc_event_t *ev)
 }
 
 void
-isc_httpdmgr_shutdown(isc_httpdmgr_t **httpdmgrp)
-{
+isc_httpdmgr_shutdown(isc_httpdmgr_t **httpdmgrp) {
 	isc_httpdmgr_t *httpdmgr;
 	isc_httpd_t *httpd;
 	httpdmgr = *httpdmgrp;
@@ -753,8 +791,7 @@ isc_httpdmgr_shutdown(isc_httpdmgr_t **httpdmgrp)
 }
 
 static isc_result_t
-grow_headerspace(isc_httpd_t *httpd)
-{
+grow_headerspace(isc_httpd_t *httpd) {
 	char *newspace;
 	unsigned int newlen;
 	isc_region_t r;
@@ -775,8 +812,7 @@ grow_headerspace(isc_httpd_t *httpd)
 }
 
 isc_result_t
-isc_httpd_response(isc_httpd_t *httpd)
-{
+isc_httpd_response(isc_httpd_t *httpd) {
 	isc_result_t result;
 	unsigned int needlen;
 
@@ -784,7 +820,7 @@ isc_httpd_response(isc_httpd_t *httpd)
 	needlen += 3 + 1;  /* room for response code, always 3 bytes */
 	needlen += strlen(httpd->retmsg) + 2;  /* return msg + CRLF */
 
-	if (isc_buffer_availablelength(&httpd->headerbuffer) < needlen) {
+	while (isc_buffer_availablelength(&httpd->headerbuffer) < needlen) {
 		result = grow_headerspace(httpd);
 		if (result != ISC_R_SUCCESS)
 			return (result);
@@ -809,7 +845,7 @@ isc_httpd_addheader(isc_httpd_t *httpd, const char *name,
 		needlen += 2 + strlen(val); /* :<space> and val */
 	needlen += 2; /* CRLF */
 
-	if (isc_buffer_availablelength(&httpd->headerbuffer) < needlen) {
+	while (isc_buffer_availablelength(&httpd->headerbuffer) < needlen) {
 		result = grow_headerspace(httpd);
 		if (result != ISC_R_SUCCESS)
 			return (result);
@@ -828,11 +864,10 @@ isc_httpd_addheader(isc_httpd_t *httpd, const char *name,
 }
 
 isc_result_t
-isc_httpd_endheaders(isc_httpd_t *httpd)
-{
+isc_httpd_endheaders(isc_httpd_t *httpd) {
 	isc_result_t result;
 
-	if (isc_buffer_availablelength(&httpd->headerbuffer) < 2) {
+	while (isc_buffer_availablelength(&httpd->headerbuffer) < 2) {
 		result = grow_headerspace(httpd);
 		if (result != ISC_R_SUCCESS)
 			return (result);
@@ -856,7 +891,7 @@ isc_httpd_addheaderuint(isc_httpd_t *httpd, const char *name, int val) {
 	needlen += 2 + strlen(buf); /* :<space> and val */
 	needlen += 2; /* CRLF */
 
-	if (isc_buffer_availablelength(&httpd->headerbuffer) < needlen) {
+	while (isc_buffer_availablelength(&httpd->headerbuffer) < needlen) {
 		result = grow_headerspace(httpd);
 		if (result != ISC_R_SUCCESS)
 			return (result);
@@ -871,11 +906,9 @@ isc_httpd_addheaderuint(isc_httpd_t *httpd, const char *name, int val) {
 }
 
 static void
-isc_httpd_senddone(isc_task_t *task, isc_event_t *ev)
-{
+isc_httpd_senddone(isc_task_t *task, isc_event_t *ev) {
 	isc_httpd_t *httpd = ev->ev_arg;
 	isc_region_t r;
-	isc_result_t result;
 	isc_socketevent_t *sev = (isc_socketevent_t *)ev;
 
 	ENTER("senddone");
@@ -898,9 +931,10 @@ isc_httpd_senddone(isc_task_t *task, isc_event_t *ev)
 	 */
 	if (httpd->freecb != NULL) {
 		isc_buffer_t *b = NULL;
-		if (isc_buffer_length(&httpd->bodybuffer) > 0)
+		if (isc_buffer_length(&httpd->bodybuffer) > 0) {
 			b = &httpd->bodybuffer;
-		httpd->freecb(b, httpd->freecb_arg);
+			httpd->freecb(b, httpd->freecb_arg);
+		}
 		NOTICE("senddone free callback performed");
 	}
 	if (ISC_LINK_LINKED(&httpd->bodybuffer, link)) {
@@ -926,8 +960,9 @@ isc_httpd_senddone(isc_task_t *task, isc_event_t *ev)
 
 	r.base = (unsigned char *)httpd->recvbuf;
 	r.length = HTTP_RECVLEN - 1;
-	result = isc_socket_recv(httpd->sock, &r, 1, task, isc_httpd_recvdone,
-				 httpd);
+	/* check return code? */
+	(void)isc_socket_recv(httpd->sock, &r, 1, task,
+			      isc_httpd_recvdone, httpd);
 
 out:
 	isc_event_free(&ev);
@@ -935,8 +970,7 @@ out:
 }
 
 static void
-reset_client(isc_httpd_t *httpd)
-{
+reset_client(isc_httpd_t *httpd) {
 	/*
 	 * Catch errors here.  We MUST be in RECV mode, and we MUST NOT have
 	 * any outstanding buffers.  If we have buffers, we have a leak.
@@ -947,6 +981,7 @@ reset_client(isc_httpd_t *httpd)
 
 	httpd->recvbuf[0] = 0;
 	httpd->recvlen = 0;
+	httpd->headers = NULL;
 	httpd->method = ISC_HTTPD_METHODUNKNOWN;
 	httpd->url = NULL;
 	httpd->querystring = NULL;
@@ -960,6 +995,14 @@ reset_client(isc_httpd_t *httpd)
 isc_result_t
 isc_httpdmgr_addurl(isc_httpdmgr_t *httpdmgr, const char *url,
 		    isc_httpdaction_t *func, void *arg)
+{
+	return (isc_httpdmgr_addurl2(httpdmgr, url, ISC_FALSE, func, arg));
+}
+
+isc_result_t
+isc_httpdmgr_addurl2(isc_httpdmgr_t *httpdmgr, const char *url,
+		     isc_boolean_t isstatic,
+		     isc_httpdaction_t *func, void *arg)
 {
 	isc_httpdurl_t *item;
 
@@ -980,6 +1023,9 @@ isc_httpdmgr_addurl(isc_httpdmgr_t *httpdmgr, const char *url,
 
 	item->action = func;
 	item->action_arg = arg;
+	item->isstatic = isstatic;
+	isc_time_now(&item->loadtime);
+
 	ISC_LINK_INIT(item, link);
 	ISC_LIST_APPEND(httpdmgr->urls, item, link);
 
